#' This job represents a sub-DAX that will be planned and executed by the workflow
#'
#' @details
#' The name argument can be either a string, or a \code{File} object. If
#' it is a \code{File} object, then this job will inherit its name from the
#' \code{File} and the \code{File} will be added in a \code{<uses>} with \code{transfer=TRUE},
#' \code{register=FALSE}, and \code{link=input}.
#'
#' @examples
#' daxjob1 <- DAX("foo.dax")
#' daxfile <- File("foo.dax")
#' daxjob2 <- DAX(daxfile)
#'
#' @param file The logical name of the DAX file or the DAX File object
#' @param id The id of the DAX job [default: autogenerated]
#' @param node.label The label for this job to use in graphing
#' @return The sub-DAX job
#' @export
DAX <- function(file, id=NULL, node.label=NULL) {

  if (class(file) == "File") {
    file = file
  } else if (is.character(file)) { # TODO: verify unicode
    file = File(name=file)
  } else {
    stop(paste("Invalid file:", file))
  }
  abstract.job <- AbstractJob(id=id, node.label=node.label)

  object <- list(abstract.job=abstract.job, file=file)
  class(object) <- "DAX"
  return(object)
}

Unicode.DAX <- function(dax) {
  return(paste("<DAX ",dax$abstract.job$id," ",dax$file$catalog.type$name,">",sep=""))
}

ToXML.DAX <- function(dax) {
  e <- Element('dax', list(
    id=dax$abstract.job$id,
    file=dax$file$catalog.type$name,
    `node-label`=dax$abstract.job$node.label
  ))
  e <- InnerXML(dax$abstract.job, e)
  return(e)
}

# ###############################
# Add-in functions for R
# ###############################

Equals.DAX <- function(dax, other) {
  if (class(other) == "DAX") {
    if (!Equals(dax$file, other$file)) {
      return(FALSE)
    }
    return(Equals(dax$abstract.job, dax$abstract.job))
  }
  return(FALSE)
}

#' @rdname AddProfile
#' @method AddProfile DAX
#' @seealso \code{\link{DAX}}
#' @export
AddProfile.DAX <- function(obj, profile) {
  obj$abstract.job$profile.mixin <- AddProfileMixin(obj$abstract.job$profile.mixin, profile)
  return(obj)
}

#' @rdname AddInvoke
#' @method AddInvoke DAX
#' @seealso \code{\link{DAX}}
#' @export
AddInvoke.DAX <- function(obj, invoke) {
  obj$abstract.job$invoke.mixin <- AddInvokeMixin(obj$abstract.job$invoke.mixin, invoke)
  return(obj)
}
